/// Type alias for json object map
typedef JsonObject = Map<String, dynamic>;

/// Converts a [JsonObject] into [T]
typedef FromJsonObject<T> = T Function(JsonObject body);

/// Converts a json element into [T]
typedef FromJson<T> = T Function(dynamic body);

/// Mixin for Data Transfer Objects
/// Every subclass of this mixin must also provide fromJson static method:
/// ```dart
/// static Foo fromJson(JsonObject json) => Foo(name: json['name'], age: json['age']);
/// ```
mixin DTO {
  /// Converts this instance into a json object map. E.g. JsonObject toJson() => {'name': name, 'age': age};
  JsonObject toJson();

  /// Default toString implementation: E.g. Foo(name: Ugo, age: 37)
  @override
  String toString() => '$runtimeType(${toJson().entries.map((entry) => '${entry.key}: ${entry.value}').join(', ')})';
}

/// Shortcut for json['xx'] == null ? null : fromJson(json['xx'])
/// Usage: fromNullableJson(json['xx'], fromJson)
T? fromNullableJson<T>(JsonObject? json, FromJsonObject<T> fromJson) => json == null ? null : fromJson(json);

/// Shortcut for (json['xx'] as List).map((json) => fromJson(json)).toList(growable: false)
/// Usage: fromJsonList(json['xx'], fromJson)
List<T> fromJsonList<T>(Iterable json, FromJsonObject<T> fromJson) =>
    json.map((json) => fromJson(json)).toList(growable: false);

/// Shortcut for (json['xx'] as List?)?.map((json) => fromJson(json)).toList(growable: false)
/// Usage: fromNullableJsonList(json['xx'], fromJson)
List<T>? fromNullableJsonList<T>(Iterable? json, FromJsonObject<T> fromJson) =>
    json?.map((json) => fromJson(json)).toList(growable: false);

extension DTOListExt<T extends DTO> on Iterable<T> {
  /// Shortcut for items.map((item) => item.toJson()).toList(growable: false)
  /// Usage: items.toJsonList()
  List<JsonObject> toJsonList() => map((item) => item.toJson()).toList(growable: false);
}
